# Derived from libmux, available in Plan 9 under /sys/src/libmux
# under the following terms:
#
# Copyright (C) 2003-2006 Russ Cox, Massachusetts Institute of Technology
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
require 'socket'

module R9P
	class Mux
		include Qlock
		include Rendez

		attr_reader :fd

		def initialize(con, opts = {}, &block)
			qlinit
			@queue = []
			@lock = self
			@outlock = Mutex.new
			@inlock = Mutex.new
			@deserialize = block
			@freetag = 0
			@wait = []
			@mwait = 0
			@nwait = 0
			@mintag = opts[:mintag] || 0
			@maxtag = opts[:maxtag] || 1<<16-1

			case con
			when IO
				@fd = con
			when String
				spec = con.split '!'
				case spec.shift
				when 'unix'
					print "#{self.class.inspect}: init UNIXSocket: #{spec[0]}\n" if $debug > 1
					@fd = UNIXSocket.new spec[0]
				when 'tcp'
					@fd = TCPSocket.new *spec
				else
					throw "Invalid connection spec"
				end
			else
				throw "Invalid connection spec"
			end
			if @fd == nil
				throw "No connection"
			end
		end

		def rpc(dat)
			r = newrpc dat

			begin
				qlock
				while @muxer && @muxer != r && r.data == nil
					r.sleep
				end

				if r.data == nil
					if @muxer && @muxer != r
						fail
					end
					@muxer = r
					qunlock
					begin
						while r.data == nil
							data = recv
							if data == nil
								qlock
								@queue.delete r
								throw "unexpected eof"
							end
							dispatch data
						end
						qlock
					ensure
						electmuxer
					end
				end

			ensure
				puttag(r)
				qunlock
			end
			r.data
		end

		def electmuxer
			@queue.each { |rpc|
				if @muxer != rpc && rpc.async == false
					@muxer = rpc
					rpc.wakeup
					return
				end
			}
			@muxer = nil
		end

		def dispatch(dat)
			tag = dat.tag - @mintag
			r = nil
			qsync {
				if tag < 0 || tag > @mwait
					printf "#{$0}: bad rpc tag: %ux\n", dat.tag
					return
				end
				r = @wait[tag]
				if r == nil || @queue.delete(r) == nil
					printf "#{$0}: bad rpc tag: %ux (noone waiting on it)\n", dat.tag
					return
				end
			}
			r.data = dat
			r.wakeup
		end

		def gettag(r)
			tag = 0

			while @nwait == @mwait
				if @mwait < @maxtag - @mintag
					mw =	if @mwait == 0
							1
						else
							@mwait << 1
						end
					@wait[mw] = nil
					@freetag = @mwait
					@mwait = mw
					break
				end
				sleep
			end

			tag = @freetag
			if @wait[tag]
				tag.upto(@mwait) { |n|
					if @wait[n] == nil
						tag = n
						break
					end
				}
				0.upto(@freetag) { |n|
					if @wait[n] == nil
						tag = n
						break
					end
				}
			end

			if @wait[tag]
				throw "nwait botch"
			end

			@nwait += 1
			@wait[tag] = r
			r.tag = tag + @mintag
			r.data.tag = r.tag
			r.data = nil
			r.tag
		end
		def puttag(r)
			t = r.tag
			if @wait[t] != r
				fail
			end
			@wait[t] = nil
			@nwait -= 1
			@freetag = t
			wakeup
		end

		def send(dat)
			dat.serialize
			len = @fd.syswrite dat.serial
			print "send(#{dat.class.inspect}): len=#{len} data=#{dat.serial.inspect}" if $debug > 1

			if len < dat.serial.length
				return false
			end
			true
		end
		def recv
			data = @fd.read 4
			if data
				len, = data.unpack "V"
				data << @fd.read(len-4)
				print "#{self.class.inspect}: recv: data=#{data.inspect}" if $debug > 1
				@deserialize.call data
			end
		end

		def newrpc(dat)
			rpc = Rpc.new self
			tag = nil

			rpc.data = dat
			qsync {
				gettag rpc
				@queue << rpc
			}

			if rpc.tag < 0 || send(dat) == false
				qsync {
					@queue.delete rpc
					puttag rpc
				}
				return nil
			end

			rpc
		end
		private :gettag, :puttag,:send, :recv, :newrpc, :electmuxer, :dispatch

		class Rpc
			include Rendez
			attr_accessor :data, :waiting, :async, :tag

			def initialize(mux)
				@lock = mux
				@mux = mux
				@data = nil
				@waiting = true
				@async = false
			end
		end
	end
end
